# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test generic initiation methods."""
import logging
from io import StringIO
import unittest
from unittest.mock import patch
from pyretis.initiation import (
    initiate_kick,
    initiate_load,
    initiate_restart,
    get_initiation_method,
)


logging.disable(logging.CRITICAL)


class TestInitiate(unittest.TestCase):
    """Run the tests for the reading external trajectories."""

    def test_get_initiation_method(self):
        """Test that we can get the method from settings."""
        cases = [
            {
                'settings': {'initial-path': {'method': 'kick'}},
                'method': initiate_kick,
            },
            {
                'settings': {'initial-path': {'method': 'KICK'}},
                'method': initiate_kick,
            },
            {
                'settings': {'initial-path': {'method': 'load'}},
                'method': initiate_load,
            },
            {
                'settings': {'initial-path': {'method': 'loAD'}},
                'method': initiate_load,
            },
            {
                'settings': {'initial-path': {'method': 'restart'}},
                'method': initiate_restart,
            },
            {
                'settings': {'initial-path': {'method': 'reSTART'}},
                'method': initiate_restart,
            },

        ]
        with patch('sys.stdout', new=StringIO()):
            for case in cases:
                method = get_initiation_method(case['settings'])
                self.assertEqual(method, case['method'])
        settings = {'initial-path': {'method': 'santas little failer'}}
        with self.assertRaises(ValueError):
            get_initiation_method(settings)


if __name__ == '__main__':
    unittest.main()



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This is a mock engine for testing external GROMACS engines.

This script is inteneded to be used as a spoof of the mdrun command
for GROMACS.
"""
import sys
import os
sys.path.insert(
    0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))
)
from time import sleep  # noqa: E402
import numpy as np  # noqa: E402
from pyretis.inout.formats.gromacs import (  # noqa: E402
    read_gromos96_file,
    write_gromos96_file,
    write_trr_frame,
)


def simple_parser(args, need_args):
    """Do some simple parsing of arguments."""
    for i, arg in enumerate(args):
        if arg in need_args:
            need_args[arg] = args[i+1]
    # Check that all options are set:
    for key, val in need_args.items():
        if val is None:
            print(f'Missing {key}', file=sys.stderr, end='\n')
            sys.exit(1)


def check_that_files_exist(keys, args):
    """Check that files in input arguments actually exist."""
    for key in keys:
        if not os.path.isfile(args[key]):
            print(f'Missing file {args[key]}',
                  file=sys.stderr, end='\n')
            sys.exit(1)


def read_mock_tpr(filename):
    """Read info from the mock tpr file."""
    config = None
    steps = 0
    nstxout = 0
    gen = False
    with open(filename, 'r', encoding='utf-8') as inputfile:
        for lines in inputfile:
            if lines.startswith('-c'):
                config = lines.split()[-1]
            elif lines.startswith('nsteps ='):
                steps = int(lines.split()[-1])
            elif lines.startswith('gen_vel ='):
                gen = lines.strip().split()[-1] == 'yes'
            elif lines.startswith('nstxout ='):
                nstxout = int(lines.strip().split()[-1])
    frame, xyz, vel, _ = read_gromos96_file(config)
    config = [frame, xyz, vel]
    return config, steps, nstxout, gen


def mock_mdrun(args, crash=False, sleep_time=None):
    """Fake the gmx mdrun command."""
    print('Running gmx mdrun, well the mock version', file=sys.stdout)
    need_args = {'-s': None, '-deffnm': None, '-c': None}
    simple_parser(args, need_args)

    check_that_files_exist(('-s',), need_args)
    config, steps, nstxout, gen = read_mock_tpr(need_args['-s'])

    print('Writing energy data...', file=sys.stdout)
    write_mock_edr(f"{need_args['-deffnm']}.edr", steps,
                   nstxout, gen=gen)

    if crash:
        print('Crash error for testing.', file=sys.stderr)
        sys.exit(1)

    print('Writing trajectory data...', file=sys.stdout)
    print(config[2][0][0], file=sys.stdout)

    write_mock_trr(f"{need_args['-deffnm']}.trr", steps, nstxout,
                   config[1], config[2], pause=sleep_time)

    if config[2][0][0] > 0:
        xyz = config[1] + steps * np.ones_like(config[1])
    else:
        xyz = config[1] - steps * np.ones_like(config[1])

    print('Writing cpt file...', file=sys.stdout)
    write_mock_cpt(f"{need_args['-deffnm']}.cpt", steps, xyz)

    print('Writing final configuration...', file=sys.stdout)
    write_gromos96_file(need_args['-c'], config[0],
                        xyz, config[2])


def write_mock_trr(filename, steps, nstxout, xyz, vel, pause=None):
    # pylint: disable=too-many-arguments
    """Write a mock TRR file."""
    xyzc = np.copy(xyz)
    velc = np.copy(vel)
    append = False
    natoms = len(xyzc)
    for i in range(0, steps + 1):
        if i % nstxout == 0:
            data = {'box': np.eye(3), 'x': xyzc, 'v': velc,
                    'natoms': natoms, 'time': i*0.002, 'step': i,
                    'lambda': 0.0}
            write_trr_frame(filename, data, append=append)
            if not append:
                append = True
            if pause:
                sleep(pause)
        if vel[0][0] > 0:
            xyzc += np.ones_like(xyz)
        else:
            xyzc -= np.ones_like(xyz)
        velc -= np.ones_like(velc)


def write_mock_cpt(filename, steps, xyz):
    """Write a mock cpt file."""
    with open(filename, 'w', encoding='utf-8') as outfile:
        outfile.write(f'Last step: {steps}\n')
        for j in xyz:
            outfile.write('{:12.7f} {:12.7f} {:12.7f}\n'.format(*j))


def read_mock_cpt(filename):
    """Read coordinates."""
    xyz = []
    steps = -1
    with open(filename, 'r', encoding='utf-8') as infile:
        for i, lines in enumerate(infile):
            if i == 0:
                steps = int(lines.strip().split()[-1])
            else:
                xyz.append([float(i) for i in lines.strip().split()])
    return steps, np.array(xyz)


def write_mock_edr(filename, steps, nstxout, gen=False, start=0):
    """Write a mock edr file."""
    header = [
        '# This file was created by the MOCK gmx program.',
        '@    title "GROMACS Energies',
        '@    xaxis  label "Time (ps)"',
        '@    yaxis  label "(kJ/mol)"',
        '@TYPE xy',
        '@ view 0.15, 0.15, 0.75, 0.85',
        '@ legend on',
        '@ legend box on',
        '@ legend loctype view',
        '@ legend 0.78, 0.8',
        '@ legend length 2',
        '@ s0 legend "Potential"',
        '@ s1 legend "Total Energy"',
        '@ s2 legend "Kinetic En."',
    ]
    fmt = '{:10.6f}  {:10.6f}  {:10.6f}  {:10.6f}\n'
    mode = 'a' if start > 0 else 'w'
    with open(filename, mode, encoding='utf-8') as output:
        if mode == 'w':
            for line in header:
                output.write(f'{line}\n')
        for i in range(0, steps + 1):
            j = float(i)
            if not gen:
                if i % nstxout == 0:
                    output.write(fmt.format(j, -j, j * 2, j))
            else:
                output.write(fmt.format(j, -j, j * 2, j + 1234.5678))


if __name__ == '__main__':
    # pylint: disable=invalid-name
    print('This is the mock mdrun command.')
    if '-crash' in sys.argv:
        mock_mdrun(sys.argv[1:], crash=True)
    else:
        if '-sleep' in sys.argv:
            mock_mdrun(sys.argv[1:], crash=False, sleep_time=0.15)
        else:
            mock_mdrun(sys.argv[1:], crash=False)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Histogram functions for data analysis.

This module defines some simple functions for histograms.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

histogram (:py:func:`.histogram`)
    Create a histogram from given data.

match_all_histograms (:py:func:`.match_all_histograms`)
    Function to match histograms, for instance from an umbrella
    sampling simulation.

histogram_and_avg (:py:func:`.histogram_and_avg`)
    Create a histogram and return bins, midpoints and simple statistics.

"""

import numpy as np

np.set_printoptions(legacy='1.21')

__all__ = ['histogram', 'match_all_histograms',
           'histogram_and_avg']


def histogram(data, bins=10, limits=(-1, 1), density=False,
              weights=None):
    """Create a histogram of the given data.

    Parameters
    ----------
    data : numpy.array
        The data for making the histogram.
    bins : int, optional
        The number of bins to divide the data into.
    limits : tuple/list, optional
        The max/min values to consider.
    density : boolean, optional
        If True the histogram will be normalized.
    weights : numpy.array, optional
        Weighting factors for data.

    Returns
    -------
    hist : numpy.array
        The binned counts.
    bins : numpy.array
        The edges of the bins.
    bin_mid : numpy.array
        The midpoint of the bins.

    """
    hist, bins = np.histogram(data, bins=bins, range=limits,
                              density=density, weights=weights)
    bin_mid = 0.5 * (bins[1:] + bins[:-1])
    return hist, bins, bin_mid


def histogram_and_avg(data, bins, density=True, weights=None):
    """Create histogram an return bins, midpoints and simple statistics.

    The simple statistics include the mean value and the standard
    deviation. The return structure is useful for plotting routines.
    The midpoints returned are the midpoints of the bins.

    Parameters
    ----------
    data : either 1D numpy.array or 2D numpy.array
        This is the data to create the histogram from.
        The eventual second dimension contains the weights.
    bins : int
        The number of bins to use for the histogram.
    density : boolean, optional
        If `density` is true, the histogram will be normalized.
    weights : numpy.array, optional
        Weighting factors for data. Not used if data contains them.

    Returns
    -------
    out[0] : numpy.array
        The histogram (frequency) values.
    out[1] : numpy.array
        The midpoints for the bins.
    out[2] : tuple of floats
        These are some simple statistics, `out[2][0]` is the average
        `out[2][1]` is the standard deviation.

    """
    local_data = data if len(data.shape) == 1 else data[:, 0]
    local_weights = weights if len(data.shape) == 1 else data[:, 1]

    hist, _, bin_mid = histogram(data, bins=bins,
                                 limits=(data.min(), data.max()),
                                 density=density,
                                 weights=weights)
    average = np.average(local_data, weights=local_weights)
    st_dev = np.sqrt(np.average((local_data-average)**2,
                                weights=local_weights))

    return hist, bin_mid, (average, st_dev)


def _match_histograms(histo1, histo2, bin_x, overlap):
    """Match two histograms given an overlapping region.

    The matching is done so that the integral of the overlapping regions
    of the two histograms are equal.

    Parameters
    ----------
    histo1 : numpy.array
        The first histogram.
    histo2 : numpy.array
        The second histogram, this is the histogram that will be scaled.
    bin_x : numpy.array
        This is the bin mid-points of the histograms. Note that we
        assume here that `histo1` and `histo2` are obtained using the
        same number of bins and limits.
    overlap : object like list, tuple or numpy.array
        This is the overlapping region.

    Returns
    -------
    out[0] : numpy.array
        A scaled version of second input histogram `histo2`.
    out[1] : float
        The calculated scale factor.

    """
    int1, int2 = 0.0, 0.0
    for histi, histj, bin_xi in zip(histo1, histo2, bin_x):
        if overlap[0] <= bin_xi < overlap[1]:
            int1 += histi
            int2 += histj
    if int2 == 0.0:
        scale_factor = 1.0
    else:
        scale_factor = int1 / int2
    return histo2 * scale_factor, scale_factor


def match_all_histograms(histograms, umbrellas):
    """Match several histograms from an umbrella sampling.

    Parameters
    ----------
    histograms : list of numpy.arrays
        The histograms to match.
    umbrellas : list of lists
        The umbrella windows used in the computation.

    Returns
    -------
    histograms_s : list of numpy.arrays
        The scaled histograms.
    scale_factor : list of floats
        The scale factors.
    matched_count : numpy.array
        Count for overall matched histogram (an "averaged" histogram).

    """
    histograms_s, scale_factor = [histograms[0][0]], [1.0]
    bin_x = histograms[0][2]
    for i in range(len(umbrellas) - 1):
        limits = (umbrellas[i+1][0], umbrellas[i][1])
        matched, scale = _match_histograms(histograms_s[-1],
                                           histograms[i+1][0], bin_x, limits)
        histograms_s.append(matched)
        scale_factor.append(scale)
    # merge histograms:
    matched_count = []
    for i, bin_xi in enumerate(bin_x):
        hist = 0.0  # histogram value at bin_xi
        norm = 0.0
        for k, umb in enumerate(umbrellas):
            if umb[0] <= bin_xi < umb[1]:
                hist += histograms_s[k][i]
                norm += 1.0
        if norm > 0.0:
            hist /= norm
        matched_count.append(hist)
    return histograms_s, scale_factor, np.array(matched_count)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Test some methods for the internal engine."""
import unittest
import numpy as np
from pyretis.engines.internal import MDEngine, VelocityVerlet, Langevin
from pyretis.core.random_gen import MockRandomGenerator
from pyretis.orderparameter.orderparameter import Position
from .test_engines import prepare_test_system


class MockEngine(MDEngine):
    """Mock MD engine for testing."""

    def __init__(self, middle):
        """Set up the engine."""
        super().__init__(1, 'Mock engine for testing')
        self.middle = middle
        self.counter = 0

    def integration_step(self, system):
        """Do some fake integration steps."""
        particles = system.particles
        if self.counter < 5:
            particles.pos = np.ones_like(particles.pos) * (self.middle - 1)
            particles.vel = np.ones_like(particles.vel) * (self.middle - 1)
        elif 5 <= self.counter < 7:
            particles.pos = np.ones_like(particles.pos) * self.middle
            particles.vel = np.ones_like(particles.vel) * self.middle
        else:
            particles.pos = np.ones_like(particles.pos) * (self.middle + 1)
            particles.vel = np.ones_like(particles.vel) * (self.middle + 1)
        self.counter += 1


class TestKick(unittest.TestCase):
    """Test Kicking methods."""

    def test_kick_across_middle(self):
        """"Test test kick_across_middle method."""
        # Test with the Velocity Verlet engine:
        engine = VelocityVerlet(0.002)
        order_function = Position(0, dim='x', periodic=False)
        rgen = MockRandomGenerator(seed=1, norm_shift=True)
        system = prepare_test_system()
        tis_settings = {
            'maxlength': 1000,
            'sigma_v': -1,
            'aimless': True,
            'zero_momentum': False,
            'rescale_energy': False,
            'allowmaxlength': False,
        }
        interface = -1.123456789
        ensemble = {'system': system,
                    'order_function': order_function,
                    'rgen': rgen}
        prev, curr = engine.kick_across_middle(ensemble, interface,
                                               tis_settings)
        order_curr1 = engine.calculate_order(ensemble)[0]
        order_curr2 = curr.particles.get_pos()[0][0]
        self.assertEqual(order_curr1, order_curr2)
        order_prev1 = prev.order
        order_prev2 = prev.particles.get_pos()[0][0]
        self.assertEqual(order_prev1, order_prev2)
        self.assertTrue(order_prev1 <= interface < order_curr1 or
                        order_curr1 < interface <= order_prev1)
        # Test with the Langevin engine:
        interface = -0.54321
        engine = Langevin(0.002, 0.3, rgen='mock', seed=1,
                          high_friction=False)
        ensemble['engine'] = engine
        prev, curr = engine.kick_across_middle(ensemble,
                                               interface,
                                               tis_settings)
        order_curr1 = engine.calculate_order(ensemble)[0]
        order_curr2 = curr.particles.get_pos()[0][0]
        self.assertEqual(order_curr1, order_curr2)
        order_prev1 = prev.order
        order_prev2 = prev.particles.get_pos()[0][0]
        self.assertEqual(order_prev1, order_prev2)
        self.assertTrue(order_prev1 <= interface < order_curr1 or
                        order_curr1 < interface <= order_prev1)
        # Test with a Mock Engine:
        interface = 3
        engine = MockEngine(interface)
        ensemble = {'system': system,
                    'order_function': order_function,
                    'engine': engine,
                    'rgen': rgen}
        prev, curr = engine.kick_across_middle(ensemble,
                                               interface,
                                               tis_settings)
        self.assertEqual(prev.order, 2)
        self.assertEqual(curr.order, 4)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Here, we just generate the initial config using PyRETIS tools."""
from pyretis.tools.lattice import generate_lattice
from pyretis.inout.format.xyz import write_xyz_file


# pylint: disable=invalid-name
filename = 'initial.xyz'
density = 9.0 / 15.0
xyz, size = generate_lattice('sq', [3, 3], density=density)
names = ['A', 'A', 'A', 'A', 'A', 'A', 'A', 'B', 'B']
print(f'Generating for density: {density}')
print(f'Box size: {size}')
print(f'Writing file: {filename}')
write_xyz_file(filename, xyz, names=names, header='Initial config')



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Method for initiating paths by loading from previous simulation(s).

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

initiate_load (:py:func:`.initiate_load`)
    A method that will get the initial path from the output of
    a previous simulation.
"""
import collections
import errno
import logging
import os
import shutil
from pyretis.core.common import compute_weight
from pyretis.core.path import Path, paste_paths
from pyretis.core.pathensemble import generate_ensemble_name
from pyretis.initiation.initiate_kick import initiate_path_ensemble_kick
from pyretis.inout import print_to_screen
from pyretis.inout.common import make_dirs, TRJ_FORMATS
from pyretis.inout.formats.order import OrderPathFile, OrderFile
from pyretis.inout.formats.energy import EnergyPathFile
from pyretis.inout.formats.path import PathIntFile, PathExtFile
from pyretis.inout.formats.xyz import read_xyz_file, convert_snapshot
from pyretis.inout.formats.gromacs import read_gromacs_generic
from pyretis.inout.formats.cp2k import read_cp2k_box
from pyretis.tools.recalculate_order import recalculate_order
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())


__all__ = ['initiate_load', 'write_order_parameters', 'clean_path',
           'read_path_files', 'read_path_files_ext', 'reorderframes',
           '_check_path', '_do_the_dirty_load_job']


def initiate_load(simulation, settings, cycle, plot_loads=False):
    """Initialise paths by loading already generated ones.

    Parameters
    ----------
    simulation : object like :py:class:`.Simulation`
        The simulation we are setting up.
    cycle : integer
        The simulation cycles we are starting at.
    settings : dictionary
        A dictionary with settings for the initiation.

    """
    folder = settings['initial-path'].get('load_folder', 'load')

    if not os.path.exists(folder):
        logger.critical('Load folder "%s" not found!', folder)
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT),
                                folder)

    # Check if there is data in the main folder
    if not os.path.exists(os.path.join(folder, 'traj.txt')):
        listoffiles = [i.name for i in os.scandir(folder) if i.is_file()]
        traj_files = [i for i in listoffiles if i[-4:] in TRJ_FORMATS]

        # If there are files, we shall compute their traj.txt and order.txt
        # to be used, if needed, later.
        if traj_files and 'order.txt' not in listoffiles:
            # This generated the traj.txt in the main load folder
            # We here borrow the 000 ensemble functions. It is a rather
            # safe assumption.
            engine = simulation.ensembles[0]['engine']
            system = simulation.ensembles[0]['system']
            order_function = simulation.ensembles[0]['order_function']
            if 'traj.txt' not in listoffiles:
                _generate_traj_txt_from_ext(folder, system,
                                            accepted=False)

            # This generated the order.txt in the main load folder
            if engine.engine_type == 'internal':
                traj = _load_trajectory(folder)
                _load_order_parameters(traj, folder, system, order_function)
            else:
                traj = _load_external_trajectory(folder, engine, copy=False)
                _load_order_parameters_ext(traj, folder, order_function,
                                           engine)

    for ensemble, set_ens in zip(simulation.ensembles, settings['ensemble']):
        path_ensemble = ensemble['path_ensemble']
        engine = ensemble['engine']
        system = ensemble['system']
        order_function = ensemble['order_function']
        name = path_ensemble.ensemble_name
        logger.info('Loading data for path path ensemble %s:', name)
        print_to_screen(f'Loading data for path path ensemble {name}')
        engine.exe_dir = path_ensemble.directory['generate']
        path = Path(simulation.rgen, maxlen=None)
        path.generated = ('re', 0, 0, 0)  # It remains for formatted loads.
        edir = os.path.join(
            folder,
            generate_ensemble_name(path_ensemble.ensemble_number))
        if not os.path.exists(os.path.join(edir, 'order.txt')):
            logger.info('Input Load missing')
            print_to_screen('No order.txt file found in %, attempt generate!',
                            edir)
            # Make sure the files are where they are supposed to be:
            _do_the_dirty_load_job(folder, edir)
            path.set_move('ld')

        if not os.path.exists(os.path.join(edir, 'traj.txt')):
            # Generate a traj.txt file for the path ensemble:
            _generate_traj_txt_from_ext(edir, system)
            path.set_move('ld')

        if engine.engine_type == 'internal':
            accept, status = read_path_files(
                path,
                edir,
                ensemble
            )
        elif engine.engine_type == 'external':
            accept, status = read_path_files_ext(
                path,
                edir,
                ensemble
            )
        else:
            raise ValueError('Unknown engine type!')

        # If a path is not acceptable, than here we try to rearrange it such
        # to satisfy each ensemble definition
        # Load paths can be accepted while still having many points outside
        # of the defined region. So even if accepted, we need to clean if the
        # path type is 'ld'.
        simtype = set_ens['simulation']['task']
        print("I am this simtype:", simtype)
        if accept and path.get_move() == 'ld':  # Some cleaning if needed
            clean_path(path, path_ensemble, simtype=simtype)
            # The path should still be checked, as something funny might occur
            accept, status = _check_path(path, path_ensemble)
            assert accept, "Path should be accepted after cleaning, but" + \
                           f" it is not! Status: {status}"
        elif set_ens['simulation']['task'] == 'explore':
            set_ens['tis']['shooting_move'] = 'exp'
            clean_path(path, path_ensemble, simtype=simtype)
            status = 'EXP'
        elif not accept:  # Let's try to fix the path
            msg = 'Sorting input files'
            logger.info(msg)
            print_to_screen(msg)
            path, (accept, status) = reorderframes(path, path_ensemble)
            clean_path(path, path_ensemble, simtype=simtype)
            path.set_move('ld')
            if not accept:
                msg = f'Path stored in load folder {edir} does not satisfy the'
                msg += f' relative ensemble ({path_ensemble.ensemble_number}) '
                msg += 'definition. \n The given path has a min of '
                msg += f'{path.ordermin[0]} and a max of {path.ordermax[0]} '
                msg += 'while the relative interface are located at '
                msg += '{}, {}, and {}.\n'.format(*path_ensemble.interfaces)
                if set_ens['initial-path'].get('load_and_kick', False):
                    msg += 'Initialising load_and_kick procedure.'
                    logger.info(msg)
                    print_to_screen(msg)
                    best_frame = path.phasepoints[0]
                    target = ensemble['path_ensemble'].interfaces[1]
                    for phasepoint in path.phasepoints:
                        if abs(best_frame.order[0] - target) > \
                                abs(phasepoint.order[0] - target):
                            best_frame = phasepoint

                    ensemble['system'] = best_frame
                    accept, path, status = initiate_path_ensemble_kick(
                        ensemble,
                        set_ens['tis'],
                        cycle)
                else:
                    msg += 'You might try to add the load_and_kick = True' \
                           ' option in the initial path section to try to fix'\
                           ' this. \n' \
                           'NOTE: points within state B are not used.'
                    raise ValueError(msg)

        path.status = status
        if set_ens['tis'].get('high_accept', False):
            move = set_ens['tis'].get('shooting_move', 'sh')
            if move == 'wf':
                intf = ensemble['path_ensemble'].interfaces
                int_cap = set_ens['tis'].get('interface_cap', intf[2])
                intf = [intf[0], intf[1], int_cap]
            if set_ens['tis'].get('shooting_move', 'sh') in ('ss', 'wf'):
                path.weight = compute_weight(path, intf, move)
        path_ensemble.add_path_data(path, status, cycle)

        # As the loading is sometimes messy, I think it's not a bad idea to
        # just plot the loaded paths, such that the user KNOWS what is loaded.
        # ignore this for coverage as it is solely for debugging purposes.
        if plot_loads:  # pragma: no cover
            left, center, right = path_ensemble.interfaces
            import matplotlib.pyplot as plt
            fig, ax = plt.subplots()
            ax.plot([p.order[0] for p in path.phasepoints], lw=1, marker='o',
                    ms=3)
            ax.axhline(y=left, ls='--', c='k')
            ax.axhline(y=center, ls='--', c='k')
            ax.axhline(y=right, ls='--', c='k')
            ax.set_xlabel('Phasepoint index')
            ax.set_ylabel('Order parameter')
            fig.savefig(f'clean_ensemble{path_ensemble.ensemble_number}.png')

        yield accept, path, status, path_ensemble


def reorderframes(path, path_ensemble):
    """Re-order the phase points in the input path.

    This function assumes that the path needs to be fixed.
    It checks for the validity of the path and keeps only the frames
    that are useful for the region to explore e.g. [0^-].

    Parameters
    ----------
    path : object like :py:class:`.PathBase`
        The path we try to fix.
    path_ensemble : object like :py:class:`.PathEnsembleExt`
        The path ensemble the path could be added to.

    """
    slave_path = path.copy()
    left, center, right = path_ensemble.interfaces
    # We first check that sorting doesn't suffice to fix the path
    path.sorting('order', reverse=True)
    accept, status = _check_path(path, path_ensemble, warning=False)
    if accept:
        return path, (accept, status)
    path.sorting('order', reverse=False)
    accept, status = _check_path(path, path_ensemble, warning=False)
    if accept:
        return path, (accept, status)

    # If the path is from L to R, we order the phase points according
    # to the order parameter.
    # If the path is from R to L (e.g. for the [0^-] ensemble),
    # a reversed path is generated.
    # If the path goes from L and stops, it is doubled to create a
    # L to L path.
    # If the path goes from R and stops, it is doubled to create a
    # R to R path.
    # NOTE: the path is now sorted.
    if 'R' == path_ensemble.start_condition and path.ordermin[0] < center:
        slave_path.sorting('order', reverse=False)
        path = paste_paths(path, slave_path)

    elif "L" == path_ensemble.start_condition and\
            path.get_end_point(left, right) not in {'L', 'R'} and\
            path.ordermax[0] >= center and path.ordermin[0] < left:
        slave_path = path.copy()
        path.sorting('order', reverse=True)
        slave_path.sorting('order', reverse=True)
        path = paste_paths(path, slave_path)
    elif set(path_ensemble.start_condition) == set(['L', 'R']):
        if path.ordermax[0] >= center and path.ordermin[0] < left:
            slave_path = path.copy()
            path.sorting('order', reverse=True)
            slave_path.sorting('order', reverse=True)
            path = paste_paths(path, slave_path)

    path.set_move('ld')

    return path, _check_path(path, path_ensemble)


def clean_path(path, path_ensemble, simtype='retis'):
    """Remove useless frames from a path.

    This function removes phasepoints from a path that are
    not needed (e.g. the one in stable states)

    Parameters
    ----------
    path : object like :py:class:`.PathBase`
        The path we try to fix.
    path_ensemble : object like :py:class:`.PathEnsembleExt`
        The path ensemble the path could be added to.
    simtype : string
        The simulation type. 'retis', 'repptis', 'pptis'

    """
    left, center, right = path_ensemble.interfaces

    # Initialize
    dists = {'left': float('inf'),
             'right': float('inf')}
    ph_min = path.phasepoints[path.ordermin[1]]
    ph_max = path.phasepoints[path.ordermax[1]]
    keep_ph = None

    # First, find the two nearest crossing points to the 0 interface.
    if path_ensemble.start_condition == "L":
        int_opt = (left, left)
        int_a, int_b = center, right
        keep_ph = path.phasepoints[path.ordermax[1]].copy()
    elif path_ensemble.start_condition == "R":
        int_opt = (center, center)
        int_a, int_b = left, center
        keep_ph = path.phasepoints[path.ordermin[1]].copy()
    else:
        int_opt = (left, right)
        int_a, int_b = left, right

    for i, phasepoint in enumerate(path.phasepoints):
        # From left, searching for closest point left of int_opt[0]
        if phasepoint.order[0] < int_opt[0]:
            if int_opt[0] - phasepoint.order[0] < dists['left']:
                dists['left'] = int_opt[0] - phasepoint.order[0]
                ph_min = phasepoint.copy()
        elif simtype == 'retis':  # RETIS behavior
            if phasepoint.order[0] - int_opt[1] < dists['right']:
                dists['right'] = phasepoint.order[0] - int_opt[1]
                ph_max = phasepoint.copy()
        elif simtype in ['pptis', 'repptis']:
            if set(["L", "R"]) == set(path_ensemble.start_condition):
                if phasepoint.order[0] > int_opt[1]:
                    if phasepoint.order[0] - int_opt[1] < dists['right']:
                        dists['right'] = phasepoint.order[0] - int_opt[1]
                        ph_max = phasepoint.copy()
            else:
                if phasepoint.order[0] - int_opt[1] < dists['right']:
                    dists['right'] = phasepoint.order[0] - int_opt[1]
                    ph_max = phasepoint.copy()
    # Prepare for deleting unnecessary frames.
    erase_list = []
    for i, phasepoint in enumerate(path.phasepoints):
        if not int_a <= phasepoint.order[0] <= int_b:
            erase_list.append(i)

    if not erase_list:
        return

    # Erase the largest elements first to avoid counting problems.
    for i in erase_list[::-1]:
        path.delete(i)

    # The path has to cope with the ensemble requirements.
    if path_ensemble.start_condition == "R":
        path.phasepoints.insert(0, ph_min.copy())
        path.phasepoints.insert(0, ph_max.copy())
        # Add the best point there is.
        if path.ordermin[0] >= center:
            path.phasepoints.append(keep_ph)
        path.phasepoints.append(ph_min.copy())
        path.phasepoints.append(ph_max.copy())

    elif path_ensemble.start_condition == 'L':
        path.phasepoints.insert(0, ph_max.copy())
        path.phasepoints.insert(0, ph_min.copy())
        # Add the best point there is.
        if path.ordermax[0] < center:
            path.phasepoints.append(keep_ph)
        path.phasepoints.append(ph_max.copy())
        path.phasepoints.append(ph_min.copy())

    else:
        path.phasepoints.insert(0, ph_min.copy())
        path.phasepoints.append(ph_max.copy())

    if simtype in ['pptis', 'repptis']:
        if ph_max.order[0] > right:
            if set(["L", "R"]) == set(path_ensemble.start_condition):
                path.sorting(key="order")
                logger.info("The load path has been sorted")


def _do_the_dirty_load_job(mainfolder, edir):
    """Copy the files to a place where PyRETIS expects them to be.

    The function checks if in the destination folder some suitable files
    are already present. If not, it tries to copy them from the main load
    folder.

    Parameters
    ----------
    mainfolder : string
        The path of the main load directory with frame/trajectory
        files.
    edir : string
        The path of the path ensemble load directory with
        frame/trajectory files.

    """
    listoffiles = [i.name for i in os.scandir(mainfolder) if i.is_file()]
    traj_files = [i for i in listoffiles if i[-4:] in TRJ_FORMATS]

    # If there is an ensemble folder, let's check what there is inside.
    if os.path.exists(edir):
        # First in the main ensemble folder
        listoffiles = [i.name for i in os.scandir(edir) if i.is_file()]
        traj_files_ens = [i for i in listoffiles if i[-4:] in TRJ_FORMATS]

        # Then in an eventual accepted folder
        traj_ens_acc = []
        if os.path.exists(os.path.join(edir, 'accepted')):
            listoffiles = [i.name for i in os.scandir(os.path.join(edir,
                                                                   'accepted'))
                           if i.is_file()]
            traj_ens_acc = [i for i in listoffiles if i[-4:] in TRJ_FORMATS]

        # If we pass the following, it means no input exist.
        if not traj_files_ens and not traj_ens_acc and not traj_files:
            logger.critical('No files to load in %s', edir)
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT),
                                    edir)

        # Here we are sure that there are inputs, we shall act being carefull
        # to not overwrite eventual info.
        # Starting from the outermost directory:
        if traj_ens_acc or traj_files_ens:
            # If there are files in the accepted folder, we just keep them.
            # If they are in the edir folder,
            # they shall be moved to the accepted folder.
            make_dirs(os.path.join(edir, 'accepted'))
            for filei in traj_files_ens:
                if filei not in {'order.txt', 'traj.txt', 'energy.txt',
                                 'pathensemble.txt'}:
                    shutil.move(
                        os.path.join(edir, filei),
                        os.path.join(edir, 'accepted', filei)
                    )
            return

    # If we arrived here, it means that we have to use the files
    # present in the main folder. So let's do the dirty copy.
    if not traj_files:
        logger.critical('No files to load in %s', mainfolder)
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT),
                                mainfolder)

    make_dirs(os.path.join(edir, 'accepted'))
    for filei in traj_files:
        if filei in {'order.txt', 'traj.txt', 'energy.txt'}:
            shutil.copy(
                os.path.join(mainfolder, filei),
                os.path.join(edir, filei)
            )
        else:
            shutil.copy(
                os.path.join(mainfolder, filei),
                os.path.join(edir, 'accepted', filei)
            )


def _generate_traj_txt_from_ext(dirname, system, accepted=True):
    """Generate a traj.txt file in the path ensemble folder.

    Parameters
    ----------
    dirname : string
        The path of the path ensemble directory with
        trajectory files.
    system : object like :py:class:`.System`
        A system object we can use when calculating the order
        parameter(s).
    accepted : boolean, optional
        It controls where the traj files shall be located,
        if True in the accepted folder, if False in the main folder.

    """
    load_file = os.path.join(dirname, 'traj.txt')

    if accepted:
        dirname = os.path.join(dirname, 'accepted')

    traj_files = sorted([i for i in os.listdir(dirname)
                         if i[-4:] in TRJ_FORMATS])

    path = Path()
    # Load Sparse works also from multiple initial trajectories
    # both internally and externally
    if system.particles.particle_type == 'internal':
        for filei in traj_files:
            if filei[-4:] == '.xyz':
                for snap in read_xyz_file(os.path.join(dirname, filei)):
                    _, pos, vel, _ = convert_snapshot(snap)
                    snapshot = {'pos': pos, 'vel': vel}
                    phase_point = system_from_snapshot(system, snapshot)
                    path.append(phase_point)
            if filei[-4:] == '.txt':
                traj = _load_trajectory(dirname, filei)
                for snapshot in traj['data']:
                    snapshot = {'pos': snapshot['pos'],
                                'vel': snapshot['vel']}
                    phase_point = system_from_snapshot(system, snapshot)
                    path.append(phase_point)

        with PathIntFile(load_file, 'w', backup=True) as pathfile:
            pathfile.output(0, (path, 'ACC'))
    else:
        for tfile in traj_files:
            if tfile[-4:] == '.txt':
                continue
            all_tfile = os.path.join(dirname, tfile)
            if tfile[-4:] == '.xyz':
                for i, snap in enumerate(read_xyz_file(all_tfile)):
                    snapshot = {'pos': (tfile, i), 'vel': False}
                    phase_point = system_from_snapshot(system, snapshot)
                    path.append(phase_point)
            else:
                for i, snap in enumerate(read_gromacs_generic(all_tfile)):
                    snapshot = {'pos': (tfile, i), 'vel': False}
                    phase_point = system_from_snapshot(system, snapshot)
                    path.append(phase_point)
        with PathExtFile(load_file, 'w', backup=True) as pathfile:
            pathfile.output(0, (path, 'ACC'))


def _load_order_parameters(traj, dirname, system, order_function):
    """Load or recalculate the order parameters.

    Parameters
    ----------
    traj : dictionary
        The trajectory we have loaded. Used here if we are
        re-calculating the order parameter(s).
    dirname : string
        The path to the directory with the input files.
    system : object like :py:class:`.System`
        A system object we can use when calculating the order
        parameter(s).
    order_function : object like :py:class:`.OrderParameter`
        This can be used to re-calculate order parameters in case
        they are not given.

    Returns
    -------
    out : list
        The order parameters, each item in the list corresponds to a
        time frame.

    """
    order_file_name = os.path.join(dirname, 'order.txt')
    try:
        with OrderPathFile(order_file_name, 'r') as orderfile:
            print_to_screen('Loading order parameters.')
            order = next(orderfile.load())
            return order['data'][:, 1:]
    except FileNotFoundError:
        print_to_screen(f'Could not read file: {order_file_name}')
    orderdata = []
    print_to_screen('Recalculating order parameters for input path.')
    logger.info('Recalculating order parameters for input path.')
    for snapshot in traj['data']:
        system.particles.pos = snapshot['pos']
        system.particles.vel = snapshot['vel']
        orderdata.append(order_function.calculate(system))
    return orderdata


def _load_order_parameters_ext(traj, dirname, order_function, engine):
    """Load or re-calculate the order parameters.

    For external trajectories, dumping of specific frames from a
    trajectory might be expensive and we here do slightly more work
    than just dumping the frames.

    Parameters
    ----------
    traj : dictionary
        The trajectory we have loaded. Used here if we are
        re-calculating the order parameter(s).
    dirname : string
        The path to the directory with the input files.
    order_function : object like :py:class:`.OrderParameter`
        This can be used to re-calculate order parameters in case
        they are not given.
    system : object like :py:class:`.System`
        A system object we can use when calculating the order
        parameter(s).

    Returns
    -------
    out : list
        The order parameters, each item in the list corresponds to
        a time frame.

    """
    order_file_name = os.path.join(dirname, 'order.txt')
    if os.path.isfile(order_file_name):
        with OrderPathFile(order_file_name, 'r') as orderfile:
            print_to_screen('Loading order parameters.')
            order = next(orderfile.load())
            return order['data'][:, 1:]
    else:
        print_to_screen('Could not read file: {order_file_name}')

    orderdata = []
    print_to_screen('Recalculating order parameters for input path.')
    logger.info('Recalculating order parameters for input path.')
    # First get unique files and indices for them:
    files = collections.OrderedDict()
    for snapshot in traj['data']:
        filename = snapshot[1]
        idx = int(snapshot[2])
        if filename not in files:
            files[filename] = {'minidx': idx-1, 'maxidx': idx+1,
                               'reverse': snapshot[3]}
        if idx < files[filename].get('minidx', idx+1):
            files[filename]['minidx'] = idx
        if idx > files[filename].get('maxidx', idx-1):
            files[filename]['maxidx'] = idx
    # Ok, now we have the files, calculate the order parameters:
    for filename, info in files.items():
        if filename[-4:] == '.xyz':           # CP2K specific NVE.
            info['box'], _ = read_cp2k_box(engine.input_files['template'])
        for order in recalculate_order(order_function, filename, info):
            orderdata.append(order)

    # Store the re-calculated order parameters so we don't have
    # to re-calculate again later:
    write_order_parameters(order_file_name, orderdata)
    return orderdata


def write_order_parameters(order_file_name, orderdata,
                           header="# Re-calculated order parameters."):
    """Store re-calculated order parameters to a file."""
    with OrderFile(order_file_name, 'w') as orderfile:
        orderfile.write(header)
        for step, data in enumerate(orderdata):
            orderfile.output(step, data)


def _load_energies_for_path(path, dirname):
    """Load energy data for a path.

    Parameters
    ----------
    path : object like :py:class:`.PathBase`
        The path we are to set up/fill.
    dirname : string
        The path to the directory with the input files.

    Returns
    -------
    None, but may add energies to the path.

    """
    # Get energies if any:
    energy_file_name = os.path.join(dirname, 'energy.txt')
    try:
        with EnergyPathFile(energy_file_name, 'r') as energyfile:
            print_to_screen('Loading energy data.')
            energy = next(energyfile.load())
            path.update_energies(energy['data']['ekin'],
                                 energy['data']['vpot'])
    except FileNotFoundError:
        print_to_screen(f'Could not read file: {energy_file_name}')
        print_to_screen('Energies are not set.')


def _check_path(path, path_ensemble, warning=True):
    """Run some checks for the path.

    Parameters
    ----------
    path : object like :py:class:`.PathBase`
        The path we are to set up/fill.
    path_ensemble : object like :py:class:`.PathEnsemble`
        The path ensemble the path could be added to.
    warning : boolean, optional
        If True, it output warnings, else only debug info.

    """
    start, end, _, cross = path.check_interfaces(path_ensemble.interfaces)
    accept = True
    status = 'ACC'
    messages = []

    if start is None or start not in path_ensemble.start_condition:
        messages.append("Initial path for %s starts at the wrong interface!")
        status = 'SWI'
        accept = False
    if end not in ('R', 'L'):
        messages.append("Initial path for %s ends at the wrong interface!")
        status = 'EWI'
        accept = False
    if not cross[1]:
        messages.append(
            "Initial path for %s does not cross the middle interface!")
        status = 'NCR'
        accept = False

    if not accept:
        msg = ' '.join(messages)
        if warning:
            logger.critical(msg, path_ensemble.ensemble_name,
                            path_ensemble.ensemble_name)
        else:
            logger.debug(msg, path_ensemble.ensemble_name,
                         path_ensemble.ensemble_name)

    path.status = status
    return accept, status


def _load_trajectory(dirname, filename='traj.txt'):
    """Load a trajectory from a file.

    Parameters
    ----------
    dirname : string
        The directory where we can find the trajectory file(s).
    filename : string
        The name of the trajectory file.

    Returns
    -------
    traj : dict
        A dictionary containing the trajectory information.
        That is, position and velocity for each frame.

    """
    with PathIntFile(os.path.join(dirname, filename), 'r') as trajfile:
        # Just get the first trajectory:
        traj = next(trajfile.load())
    return traj


def read_path_files(path, dirname, ensemble):
    """Read data needed for a path from a directory.

    Parameters
    ----------
    path : object like :py:class:`.PathBase`
        The path we are to set up/fill.
    dirname : string
        The path to the directory with the input files.
    ensemble : dictionary of objects
        It contains:

        * `path_ensemble`: object like :py:class:`.PathEnsemble`
          This is the path ensemble to perform the initialization.
        * `order_function`: object like :py:class:`.OrderParameter`
          The class used for obtaining the order parameter(s).
        * `engine`: object like :py:class:`.EngineBase`
          The engine to use for propagating a path.
        * `system`: object like :py:class:`.System`
          A system object we can use when calculating the order parameter(s).

    """
    path_ensemble = ensemble['path_ensemble']
    engine = ensemble['engine']
    system = ensemble['system']
    order_function = ensemble['order_function']
    left, _, right = path_ensemble.interfaces
    traj = _load_trajectory(dirname)
    orderdata = _load_order_parameters(traj, dirname, system, order_function)

    # Add to path:
    print_to_screen('Creating path from files.')
    logger.debug('Creating path from files.')
    for snapshot, orderi in zip(traj['data'], orderdata):
        snapshot = {'order': orderi, 'pos': snapshot['pos'],
                    'vel': snapshot['vel']}
        phase_point = system_from_snapshot(system, snapshot)
        engine.add_to_path(
            path,
            phase_point,
            left,
            right
        )
    _load_energies_for_path(path, dirname)
    return _check_path(path, path_ensemble)


def _load_external_trajectory(dirname, engine, copy=True):
    """Load an external trajectory.

    Here, we also do some moving of files to set up for a path
    simulation (if copy=True).

    Parameters
    ----------
    dirname : string
        The directory where we can find the trajectory file(s).
    engine : object like :py:class:`.ExternalMDEngine`
        The engine we use, here it is used to access the directories
        for the new simulation.
    copy : boolean, opt
        If true, it copies the files to the main ensemble folders.

    Returns
    -------
    traj : dict
        A dictionary containing the trajectory information. Here,
        the trajectory information is a list of files with indices and
        information about velocity direction.

    """
    traj_file_name = os.path.join(dirname, 'traj.txt')
    with PathExtFile(traj_file_name, 'r') as trajfile:
        # Just get the first trajectory:
        traj = next(trajfile.load())
        # Hard-copy the files. Here we assume that they are stored
        # in an folder called "accepted" and that we can get the
        # **correct** names from the traj.txt file!
        files = set([])
        for snapshot in traj['data']:
            filename = os.path.join(os.path.abspath(dirname),
                                    'accepted', snapshot[1])
            files.add(filename)
        print_to_screen('Copying trajectory files.')

        # If the files are not copied, the location has to be consistent (loc)
        loc = dirname
        if copy:
            loc = engine.exe_dir
            for filename in files:
                logger.debug('Copying %s -> %s', filename, engine.exe_dir)
                shutil.copy(filename, engine.exe_dir)
        # Update trajectory to use full path names:
        for i, snapshot in enumerate(traj['data']):
            config = os.path.join(loc, snapshot[1])
            traj['data'][i][1] = config
            reverse = int(snapshot[3]) == -1
            idx = int(snapshot[2])
            traj['data'][i][2] = idx
            traj['data'][i][3] = reverse
        return traj


def read_path_files_ext(path, dirname, ensemble):
    """Read data needed for a path from a directory.

    Parameters
    ----------
    path : object like :py:class:`.Path`
        The path we are to set up/fill.
    path_ensemble : object like :py:class:`.PathEnsembleExt`
        The path ensemble the path could be added to.
    dirname : string
        The path to the directory with the input files.
    ensemble : dictionary of objects
        It contains:

        * `path_ensemble`: object like :py:class:`.PathEnsemble`
          This is the path ensemble to perform the initialization.
        * `order_function`: object like :py:class:`.OrderParameter`
          The class used for obtaining the order parameter(s).
        * `engine`: object like :py:class:`.EngineBase`
          The engine to use for propagating a path.
        * `system`: object like :py:class:`.System`
          A system object we can use when calculating the order parameter(s).

    """
    path_ensemble = ensemble['path_ensemble']
    engine = ensemble['engine']
    system = ensemble['system']
    order_function = ensemble['order_function']
    left, _, right = path_ensemble.interfaces
    traj = _load_external_trajectory(dirname, engine)
    orderdata = _load_order_parameters_ext(traj, dirname,
                                           order_function, engine)
    # Add to path:
    print_to_screen('Creating path from files.')
    logger.debug('Creating path from files.')
    for snapshot, order in zip(traj['data'], orderdata):
        snapshot = {'order': order,
                    'pos': (snapshot[1], snapshot[2]),
                    'vel': snapshot[3],
                    'vpot': None,
                    'ekin': None}
        phase_point = system_from_snapshot(system, snapshot)
        engine.add_to_path(
            path,
            phase_point,
            left,
            right
        )
    _load_energies_for_path(path, dirname)
    return _check_path(path, path_ensemble)


def system_from_snapshot(system, snapshot):
    """Create a system from a given snapshot."""
    system_copy = system.copy()
    system_copy.particles.ekin = snapshot.get('ekin', None)
    system_copy.particles.vpot = snapshot.get('vpot', None)
    system_copy.order = snapshot.get('order', None)
    system_copy.particles.set_pos(snapshot.get('pos', None))
    system_copy.particles.set_vel(snapshot.get('vel', None))
    return system_copy



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Definition of simulation objects for Monte Carlo simulations.

This module defines some classes and functions for performing
Monte Carlo simulations.

Important classes defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

UmbrellaWindowSimulation (:py:class:`.UmbrellaWindowSimulation`)
    Defines a simulation for performing umbrella window simulations.
    Several umbrella window simulations can be joined to perform a
    umbrella simulation.
"""
import numpy as np
from pyretis.core.montecarlo import max_displace_step
from pyretis.simulation.simulation import Simulation
from pyretis.core.random_gen import create_random_generator

__all__ = ['UmbrellaWindowSimulation']


def mc_task(rgen, system, maxdx):
    """Perform a Monte Carlo displacement move.

    Here, a displacement step will be carried out and
    the trial move will be accepted/rejected and the
    positions and potential energy will be updated as needed.

    Parameters
    ----------
    rgen : object like :py:class`.RandomGenerator`
        This object is used for generating random numbers.
    system : object like :py:class:`.System`
        The system we act on.
    maxdx : float
        Maximum displacement step for the Monte Carlo move.

    """
    accepted_r, _, trial_r, v_trial, status = max_displace_step(rgen, system,
                                                                maxdx)
    if status:
        system.particles.pos = accepted_r
        system.particles.vpot = v_trial
    return accepted_r, v_trial, trial_r, status


class UmbrellaWindowSimulation(Simulation):
    """This class defines an Umbrella simulation.

    The Umbrella simulation is a special case of
    the simulation class with settings to simplify the
    execution of the umbrella simulation.

    Attributes
    ----------
    ensemble : dict
        It contains the simulations info
          *  system : object like :py:class:`.System`
                 The system to act on.
          *  rgen : object like :py:class:`.RandomGenerator`
                 Object to use for random number generation.
    umbrella : list, [float, float]
        The umbrella window to consider.
    overlap : float
        The position we have to cross before the simulation ends.
    maxdx : float
        Defines the maximum movement allowed in the Monte Carlo
    steps : int, optional
        The number of simulation steps to perform.
    startcycle : int, optional
        The cycle we start the simulation on, can be useful if
        restarting.

    """

    simulation_type = 'umbrella-window'

    def __init__(self, ensemble, settings, controls=None):
        """Initialise the umbrella simulation simulation.

        Parameters
        ----------
        ensemble : dict
            It contains the simulations info
              *  system : object like :py:class:`.System`
                     The system to act on.
              *  rgen : object like :py:class:`.RandomGenerator`
                     Object to use for random number generation.
        settings : dict
            Contains all the simulation settings.
        controls: dict of parameters to control the simulations. Optional
            It can contain:
              *  mincycle : int, optional
                     The *MINIMUM* number of cycles to perform. Note that in
                     base `Simulation` class this is the *MAXIMUM* number of
                     cycles to perform. The meaning is redefined in this class
                     by overriding `self.simulation_finished`.
              *  startcycle : int, optional
                     The current simulation cycle, i.e. where we start.

        """
        super().__init__(settings, controls)
        self.umbrella = settings['simulation']['umbrella']
        self.overlap = settings['simulation']['overlap']
        self.maxdx = settings['simulation']['maxdx']
        self.rgen = ensemble.get(
            'rgen',
            create_random_generator(settings['simulation']))

        self.system = ensemble['system']
        task_monte_carlo = {'func': mc_task,
                            'args': [self.rgen, self.system, self.maxdx],
                            'result': 'displace_step'}
        self.add_task(task_monte_carlo)
        self.first_step = False

        if 'restart' in settings:
            self.load_restart_info(settings['restart'])

    def is_finished(self):
        """Check if the simulation is done.

        In the umbrella simulation, the simulation is finished when we
        cycle is larger than maxcycle and all particles have
        crossed self.overlap.

        Returns
        -------
        out : boolean
            True if the simulation is finished, False otherwise.

        """
        return (self.cycle['step'] > self.cycle['endcycle'] and
                np.all(self.system.particles.pos > self.overlap))

    def __str__(self):
        """Return some info about the simulation as a string."""
        msg = ['Umbrella window simulation']
        msg += [f'Umbrella: {self.umbrella}, Overlap: {self.overlap}.']
        msg += [f"Minimum number of cycles: {self.cycle['endcycle']}"]
        return '\n'.join(msg)

    def restart_info(self):
        """Return information which can be used to restart the simulation.

        Returns
        -------
        info : dict,
            Contains all the updated simulation settings and counters.

        """
        info = super().restart_info()

        info['simulation'].update(self.rgen.get_state())
        info['umbrella'] = self.umbrella
        info['overlap'] = self.overlap
        info['type'] = self.simulation_type
        return info

    def load_restart_info(self, info):
        """Load the restart information."""
        super().load_restart_info(info)

        self.umbrella = info.get('umbrella', self.umbrella)
        self.overlap = info.get('overlap', self.overlap)
        self.rgen = create_random_generator(info['simulation'])



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A script for building the C Lennard-Jones extension."""
from setuptools import (
    setup,
    Extension,
)
import numpy as np


LJMODULE = Extension(
    'ljc',
    sources=['ljc.c'],
    extra_compile_args=["-Ofast", "-march=native"]
)
setup(
    name="PyRETIS Lennard-Jones C extension",
    description="C extension for the Lennard-Jones potential.",
    ext_modules=[LJMODULE],
    include_dirs=[np.get_include()],
)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""
Update the number of steps for a GROMACS input mdp file.
"""
import sys
from pyretis.inout.settings import parse_settings_file


def main(settings_file, mdp_file, output_file):
    """Update number of steps in the gromacs input file."""
    settings = parse_settings_file(settings_file)
    steps = settings['simulation']['steps']
    subcycles = settings['engine']['subcycles']
    gmx_steps = steps * subcycles
    with open(mdp_file, 'r') as infile:
        with open(output_file, 'w') as output:
            for lines in infile:
                if lines.strip().startswith('nsteps'):
                    output.write(f'nsteps = {gmx_steps}\n')
                else:
                    output.write(lines)


if __name__ == '__main__':
    main(sys.argv[1], sys.argv[2], sys.argv[3])



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""A test module for the GROMACS in/out module."""
import logging
import unittest
import tempfile
import os
import struct
import numpy as np
from pyretis.core.box import box_matrix_to_list
from pyretis.inout.formats.gromacs import (
    _GROMACS_MAGIC,
    _TRR_VERSION_B,
    read_gromacs_lines,
    read_gromacs_file,
    read_gromacs_gro_file,
    write_gromacs_gro_file,
    read_gromos96_file,
    write_gromos96_file,
    read_xvg_file,
    swap_integer,
    swap_endian,
    read_trr_file,
    read_trr_frame,
    reverse_trr,
    trr_frame_to_g96,
    write_trr_frame,
)


logging.disable(logging.CRITICAL)
HERE = os.path.abspath(os.path.dirname(__file__))


RAW_FILE_VEL = [
    'Example from: https://en.wikipedia.org/wiki/XYZ_file_format',
    '5',
    '    1DUM     Ba    1   0.000   0.000   0.000   1.000   1.000   1.000',
    '    2DUM     Hf    2   0.050   0.050   0.050  -1.000  -1.000  -1.000',
    '    3DUM      O    3   0.050   0.050   0.000   2.000   0.000  -2.000',
    '    3DUM      O    4   0.050   0.000   0.050  -2.000   1.000   2.000',
    '    3DUM      O    5   0.000   0.050   0.050   0.000  -1.000   0.000',
    '   2.00000   2.00000   2.00000',
]

RAW_FILE = [
    'Example from: https://en.wikipedia.org/wiki/XYZ_file_format',
    '5',
    '    1DUM     Ba    1   0.000   0.000   0.000',
    '    2DUM     Hf    2   0.050   0.050   0.050',
    '    3DUM      O    3   0.050   0.050   0.000',
    '    3DUM      O    4   0.050   0.000   0.050',
    '    3DUM      O    5   0.000   0.050   0.050',
    '   2.00000   2.00000   2.00000',
]

CORRECT_GRO_VEL = {
    'x': [0.0, 0.05, 0.05, 0.05, 0.0],
    'y': [0.0, 0.05, 0.05, 0.0, 0.05],
    'z': [0.0, 0.05, 0.0, 0.05, 0.05],
    'vx': [1.0, -1.0, 2.0, -2.0, 0.0],
    'vy': [1.0, -1.0, 0.0, 1.0, -1.0],
    'vz': [1.0, -1.0, -2.0, 2.0, 0.0],
    'residunr': [1, 2, 3, 3, 3],
    'atomnr': [1, 2, 3, 4, 5],
    'header': RAW_FILE[0],
    'residuname': ['DUM', 'DUM', 'DUM', 'DUM', 'DUM'],
    'atomname': ['Ba', 'Hf', 'O', 'O', 'O'],
    'box': np.array([2.0, 2.0, 2.0]),
    'xyz': np.transpose(np.array([[0.0, 0.05, 0.05, 0.05, 0.0],
                                  [0.0, 0.05, 0.05, 0.0, 0.05],
                                  [0.0, 0.05, 0.0, 0.05, 0.05]])),
    'vel': np.transpose(np.array([[1.0, -1.0, 2.0, -2.0, 0.0],
                                  [1.0, -1.0, 0.0, 1.0, -1.0],
                                  [1.0, -1.0, -2.0, 2.0, 0.0]])),
}


CORRECT_GRO_NOVEL = {
    'x': [0.0, 0.05, 0.05, 0.05, 0.0],
    'y': [0.0, 0.05, 0.05, 0.0, 0.05],
    'z': [0.0, 0.05, 0.0, 0.05, 0.05],
    'residunr': [1, 2, 3, 3, 3],
    'atomnr': [1, 2, 3, 4, 5],
    'header': RAW_FILE[0],
    'residuname': ['DUM', 'DUM', 'DUM', 'DUM', 'DUM'],
    'atomname': ['Ba', 'Hf', 'O', 'O', 'O'],
    'box': np.array([2.0, 2.0, 2.0]),
    'xyz': np.transpose(np.array([[0.0, 0.05, 0.05, 0.05, 0.0],
                                  [0.0, 0.05, 0.05, 0.0, 0.05],
                                  [0.0, 0.05, 0.0, 0.05, 0.05]])),
    'vel': np.zeros((5, 3)),
}


CORRECT_GRO_NOVEL1D = {
    'x': [0.0, 0.05, 0.05, 0.05, 0.0],
    'residunr': [1, 2, 3, 3, 3],
    'atomnr': [1, 2, 3, 4, 5],
    'header': RAW_FILE[0],
    'residuname': ['DUM', 'DUM', 'DUM', 'DUM', 'DUM'],
    'atomname': ['Ba', 'Hf', 'O', 'O', 'O'],
    'box': np.array([2.0, 2.0, 2.0]),
    'xyz': np.transpose(np.array([[0.0, 0.05, 0.05, 0.05, 0.0],
                                  [0.0, 0.0, 0.0, 0.0, 0.0],
                                  [0.0, 0.0, 0.0, 0.0, 0.0]])),
    'vel': np.zeros((5, 3)),
}


CORRECT_GRO_NOVEL2D = {
    'x': [0.0, 0.05, 0.05, 0.05, 0.0],
    'y': [0.0, 0.05, 0.05, 0.0, 0.05],
    'residunr': [1, 2, 3, 3, 3],
    'atomnr': [1, 2, 3, 4, 5],
    'header': RAW_FILE[0],
    'residuname': ['DUM', 'DUM', 'DUM', 'DUM', 'DUM'],
    'atomname': ['Ba', 'Hf', 'O', 'O', 'O'],
    'box': np.array([2.0, 2.0, 2.0]),
    'xyz': np.transpose(np.array([[0.0, 0.05, 0.05, 0.05, 0.0],
                                  [0.0, 0.05, 0.05, 0.0, 0.05],
                                  [0.0, 0.0, 0.0, 0.0, 0.0]])),
    'vel': np.zeros((5, 3)),
}


CORRECT_XVG = {
    'potential': np.array([-937898.125, -965421.6875, -992991.0,
                           -1015228.375, -1037830.125, -1046103.3125]),
    'pressure': np.array([-22567.054688, -22762.349609, -22956.455078,
                          -23085.162109, -23215.666016, -23260.259766]),
    't-rest': np.array([0.0] * 6),
    'step': np.array([i for i in range(6)]),
}


def generate_trr_data(output, steps, natoms, double=False, endian=None):
    """Generate some random TRR data."""
    all_data = []
    for i in range(steps):
        data = {
            'natoms': natoms,
            'step': i,
            'time': 0.002*i,
            'lambda': 0.0,
            'box': np.random.ranf(size=(3, 3)),
            'x': np.random.ranf(size=(natoms, 3)),
            'v': np.random.ranf(size=(natoms, 3)),
        }
        header = write_trr_frame(output, data, double=double, append=True,
                                 endian=endian)
        all_data.append((header, data))
    return all_data


class TestGromacsIO(unittest.TestCase):
    """Test Gromacs input output."""

    @staticmethod
    def assert_equal_snapshot(snapi, snapj):
        """Compare two snapshots."""
        for key, _ in snapi.items():
            if key not in snapj:
                msg = ('Snapshots have different keys'
                       ' {} is in first but not in second').format(key)
                raise AssertionError(msg)

        for key, _ in snapj.items():
            if key not in snapi:
                msg = ('Snapshots have different keys'
                       ' {} is in second but not in first').format(key)
                raise AssertionError(msg)

        for key, val in snapi.items():
            if key in ('xyz', 'vel'):
                if not np.allclose(val, snapj[key]):
                    raise AssertionError(
                        f'Snapshots not equal. "{key}" differ'
                    )
            else:
                for i, j in zip(val, snapj[key]):
                    if not i == j:
                        raise AssertionError(
                            f'Snapshots not equal. "{key}" differ'
                        )

    def test_read_gromacs_lines(self):
        """Test that we can read GROMACS lines."""
        for snapshot, correct in ((RAW_FILE_VEL, CORRECT_GRO_VEL),
                                  (RAW_FILE, CORRECT_GRO_NOVEL)):
            snap = next(read_gromacs_lines(snapshot))
            self.assert_equal_snapshot(snap, correct)
        multi = [i for i in RAW_FILE_VEL] * 2
        for snap in read_gromacs_lines(multi):
            self.assert_equal_snapshot(snap, CORRECT_GRO_VEL)

    def test_read_gromacs_file(self):
        """Test that we can read a GROMACS file."""
        filename = os.path.join(HERE, 'config.gro')
        for snap in read_gromacs_file(filename):
            self.assert_equal_snapshot(snap, CORRECT_GRO_VEL)

    def test_read_gromacs_file2(self):
        """Test that we can read a single config GROMACS file."""
        for name, correct in (('config.gro', CORRECT_GRO_VEL),
                              ('config-novel.gro', CORRECT_GRO_NOVEL)):
            filename = os.path.join(HERE, name)
            frame, xyz, vel, box = read_gromacs_gro_file(filename)
            self.assertIsNotNone(xyz)
            self.assertIsNotNone(vel)
            self.assertIsNotNone(box)
            self.assert_equal_snapshot(frame, correct)
        # Test what we get if we read 1D and 2D files:
        for dim, correct in ((1, CORRECT_GRO_NOVEL1D),
                             (2, CORRECT_GRO_NOVEL2D)):
            filename = os.path.join(HERE, f'config-{dim}D.gro')
            frame, xyz, vel, box = read_gromacs_gro_file(filename)
            self.assertIsNotNone(xyz)
            self.assertIsNotNone(vel)
            self.assertIsNotNone(box)
            self.assert_equal_snapshot(frame, correct)

    def test_write_gromacs_gro(self):
        """Test that we can write GROMACS GRO files."""
        filename = os.path.join(HERE, 'config.gro')
        frame, xyz, vel, _ = read_gromacs_gro_file(filename)
        with tempfile.NamedTemporaryFile() as tmp:
            write_gromacs_gro_file(tmp.name, frame, xyz, vel)
            tmp.flush()
            frame2, _, _, _ = read_gromacs_gro_file(tmp.name)
            self.assert_equal_snapshot(frame, frame2)
            write_gromacs_gro_file(tmp.name, frame, xyz)
            tmp.flush()
            _, xyz1, _, _ = read_gromacs_gro_file(tmp.name)
            self.assertTrue(np.allclose(xyz, xyz1))

    def test_read_gromosg96(self):
        """Test that we can read GROMACS g96 files."""
        for name in ('config.g96', 'config-novel.g96', 'config-red.g96'):
            filename = os.path.join(HERE, name)
            frame, xyz, vel, box = read_gromos96_file(filename)
            self.assertTrue(np.allclose(box, np.array(CORRECT_GRO_VEL['box'])))
            self.assertTrue(np.allclose(xyz, CORRECT_GRO_VEL['xyz']))
            if frame['VELOCITY']:
                self.assertTrue(np.allclose(vel, CORRECT_GRO_VEL['vel']))
            else:
                self.assertTrue(np.allclose(vel, np.zeros_like(vel)))

    def test_write_gromos96(self):
        """Test that we can write GROMACS g96 files."""
        filename = os.path.join(HERE, 'config.g96')
        frame, xyz, vel, box = read_gromos96_file(filename)
        with tempfile.NamedTemporaryFile() as tmp:
            write_gromos96_file(tmp.name, frame, xyz, vel, box)
            tmp.flush()
            frame2, xyz2, vel2, box2 = read_gromos96_file(tmp.name)
            self.assertTrue(np.allclose(xyz, xyz2))
            self.assertTrue(np.allclose(vel, vel2))
            self.assertTrue(np.allclose(box, box2))
            for key, val in frame.items():
                self.assertEqual(val, frame2[key])
            box = [0]*9
            write_gromos96_file(tmp.name, frame, xyz, vel, box)
            tmp.flush()
            _, _, _, box2 = read_gromos96_file(tmp.name)
            self.assertTrue(np.allclose(box, box2))

        # Test that we can write without some fields:
        del frame['VELOCITY']
        del frame['TITLE']
        del frame['BOX']
        with tempfile.NamedTemporaryFile() as tmp:
            write_gromos96_file(tmp.name, frame, xyz, vel)
            tmp.flush()
            frame2, xyz2, vel2, box2 = read_gromos96_file(tmp.name)
            self.assertEqual(None, box2)
            self.assertTrue(np.allclose(vel2, np.zeros_like(xyz2)))
            self.assertEqual(len(frame2['TITLE']), 0)
            self.assertTrue(np.allclose(xyz, xyz2))

    def test_read_xvg_file(self):
        """Test that we can read GROMACS xvg files."""
        filename = os.path.join(HERE, 'energy.xvg')
        data = read_xvg_file(filename)
        for key, val in data.items():
            self.assertTrue(np.allclose(val, CORRECT_XVG[key]))


class TestGromacsTRR(unittest.TestCase):
    """Test GROMACS TRR input output."""

    def test_swap_integer(self):
        """Test the swap_integer method."""
        test = [(1, 16777216), (2, 33554432), (4, 67108864),
                (8, 134217728), (16, 268435456)]
        for i, j in test:
            self.assertEqual(i, swap_integer(j))
            self.assertEqual(j, swap_integer(i))

    def test_swap_endian(self):
        """Test the swap_endian method."""
        test = [('>', '<'), ('<', '>')]
        for i, j in test:
            self.assertEqual(j, swap_endian(i))
        with self.assertRaises(ValueError):
            swap_endian(1)
        with self.assertRaises(ValueError):
            swap_endian('^')

    def test_read_trr(self):
        """Test the reading of TRR files."""
        filename = os.path.join(HERE, 'traj.trr')
        all_data = []
        for i, (header, data) in enumerate(read_trr_file(filename)):
            self.assertEqual(i * 10, header['step'])
            self.assertEqual(16, header['natoms'])
            all_data.append(data)
        all_data2 = []
        for i in range(11):
            _, data = read_trr_frame(filename, i)
            all_data2.append(data)
        self.assertEqual(len(all_data), len(all_data2))
        for data1, data2 in zip(all_data, all_data2):
            for key, val in data1.items():
                self.assertTrue(np.allclose(val, data2[key]))
        header, data = read_trr_frame(filename, 100)
        self.assertTrue(header is None)
        self.assertTrue(data is None)
        header, data = read_trr_frame(filename, -1)
        for i, (header, data) in enumerate(read_trr_file(filename,
                                                         read_data=False)):
            self.assertTrue(data is None)
            self.assertEqual(i * 10, header['step'])

        filename = os.path.join(HERE, 'error.trr')
        logging.disable(logging.INFO)

        with self.assertLogs('pyretis.inout.formats.gromacs',
                             level='WARNING'):
            for i in read_trr_file(filename):
                pass

        with self.assertLogs('pyretis.inout.formats.gromacs',
                             level='CRITICAL'):
            for i in read_trr_file(filename):
                pass
        logging.disable(logging.CRITICAL)

    def test_reverse_trr(self):
        """Test that we can reverse a TRR file."""
        filename = os.path.join(HERE, 'traj.trr')
        with tempfile.NamedTemporaryFile() as tmp:
            reverse_trr(filename, tmp.name, print_progress=False)
            tmp.flush()
            rev_header = []
            rev_data = []
            for header, data in read_trr_file(tmp.name):
                rev_header.append(header)
                rev_data.append(data)
            all_header = []
            all_data = []
            for header, data in read_trr_file(filename):
                all_header.append(header)
                all_data.append(data)
            for i, j in zip(all_header, reversed(rev_header)):
                for key, val in i.items():
                    self.assertEqual(val, j[key])
            for i, j in zip(all_data, reversed(rev_data)):
                for key, val in i.items():
                    self.assertTrue(np.allclose(val, j[key]))

    def test_trr_to_g96(self):
        """Test that we can extract a GROMOS g96 frame from a TRR."""
        for fname in ('traj.trr', 'traj-tric.trr'):
            filename = os.path.join(HERE, fname)
            with tempfile.NamedTemporaryFile() as tmp:
                trr_frame_to_g96(filename, 2, tmp.name)
                tmp.flush()
                _, xyz, vel, box = read_gromos96_file(tmp.name)
                _, data = read_trr_frame(filename, 2)
                self.assertTrue(np.allclose(xyz, data['x']))
                self.assertTrue(np.allclose(vel, data['v']))
                self.assertTrue(
                    np.allclose(box,
                                box_matrix_to_list(data['box'], full=True))
                )

    def test_read_double_trr(self):
        """Test that we can read double precision TRR as well."""
        filename1 = os.path.join(HERE, 'traj-double.trr')
        filename2 = os.path.join(HERE, 'traj-single.trr')
        for double, single in zip(read_trr_file(filename1),
                                  read_trr_file(filename2)):
            self.assertAlmostEqual(double[0]['time'], single[0]['time'],
                                   places=5)
            self.assertTrue(double[0]['double'])
            self.assertFalse(single[0]['double'])

    def test_write_trr(self):
        """Test that we can write simple TRR files."""
        compare_direct = {'natoms', 'vir_size', 'ir_size', 'sym_size',
                          'top_size', 'v_size', 'f_size', 'box_size',
                          'x_size', 'step', 'pres_size', 'nre', 'e_size',
                          'double'}
        cases = (
            {'double': False},
            {'double': True},
            {'double': False, 'endian': '<'},
            {'double': False, 'endian': '>'},
        )
        for case in cases:
            with tempfile.NamedTemporaryFile() as tmp:
                all_data = generate_trr_data(tmp.name, 10, 11,
                                             double=case.get('double', False),
                                             endian=case.get('endian', None))
                tmp.flush()
                for frame, correct in zip(read_trr_file(tmp.name), all_data):
                    header, data = frame
                    header2, data2 = correct
                    for key in compare_direct:
                        self.assertEqual(header[key], header2[key])
                    for key in ('time', 'lambda'):
                        self.assertAlmostEqual(header[key], header2[key])
                    for key, val in data.items():
                        self.assertTrue(np.allclose(val, data2[key]))
                    if header2['endian']:
                        self.assertEqual(header['endian'], header2['endian'])

    def test_read_wrong_header(self):
        """Test that we get an error when reading wrong version of TRR."""
        slen = (13, 12)
        fmt = ['1i', '2i', f'{slen[0] - 1}s', '13i']
        with tempfile.NamedTemporaryFile() as tmp:
            with open(tmp.name, 'wb') as outfile:
                outfile.write(struct.pack(fmt[0], _GROMACS_MAGIC))
                outfile.write(struct.pack(fmt[1], *slen))
                outfile.write(struct.pack(fmt[2], b'NOT_GMX_FILE'))
                head = [0, 0, 26, 0, 0, 0, 0, 1000, 0, 0, 10, 0, 0]
                outfile.write(struct.pack(fmt[3], *head))
                outfile.write(struct.pack('1f', 0.0))
                outfile.write(struct.pack('1f', 0.0))
            with self.assertRaises(ValueError):
                for _ in read_trr_file(tmp.name):
                    pass

    def test_read_size(self):
        """Test that we can get double/float when box is missing."""
        slen = (13, 12)
        fmt = ['1i', '2i', f'{slen[0] - 1}s', '13i']
        with tempfile.NamedTemporaryFile() as tmp:
            with open(tmp.name, 'wb') as outfile:
                outfile.write(struct.pack(fmt[0], _GROMACS_MAGIC))
                outfile.write(struct.pack(fmt[1], *slen))
                outfile.write(struct.pack(fmt[2], _TRR_VERSION_B))
                x_size = 3 * 10 * struct.calcsize('f')
                head = [0, 0, 0, 0, 0, 0, 0, x_size, 0, 0, 10, 0, 0]
                outfile.write(struct.pack(fmt[3], *head))
                outfile.write(struct.pack('1f', 0.0))
                outfile.write(struct.pack('1f', 0.0))
            for frame, _ in read_trr_file(tmp.name, read_data=False):
                self.assertFalse(frame['double'])

    def test_read_size_fail(self):
        """Test that we fail when we can't find precision."""
        slen = (13, 12)
        fmt = ['1i', '2i', f'{slen[0] - 1}s', '13i']
        with tempfile.NamedTemporaryFile() as tmp:
            with open(tmp.name, 'wb') as outfile:
                outfile.write(struct.pack(fmt[0], _GROMACS_MAGIC))
                outfile.write(struct.pack(fmt[1], *slen))
                outfile.write(struct.pack(fmt[2], _TRR_VERSION_B))
                x_size = 3 * 10 * (struct.calcsize('d') + 1)
                head = [0, 0, 0, 0, 0, 0, 0, x_size, 0, 0, 10, 0, 0]
                outfile.write(struct.pack(fmt[3], *head))
                outfile.write(struct.pack('1f', 0.0))
                outfile.write(struct.pack('1f', 0.0))
            with self.assertRaises(ValueError):
                for _ in read_trr_file(tmp.name, read_data=False):
                    pass

    def test_overwrite_trr(self):
        """Test that we indeed can turn off the append to TRR."""
        with tempfile.NamedTemporaryFile() as tmp:
            all_data = []
            for i in range(5):
                data = {
                    'natoms': 4,
                    'step': i,
                    'time': 0.002*i,
                    'lambda': 0.0,
                    'box': np.random.ranf(size=(3, 3)),
                    'x': np.random.ranf(size=(4, 3)),
                }
                header = write_trr_frame(tmp.name, data, double=False,
                                         append=False)
                all_data.append((header, data))
            for frame in read_trr_file(tmp.name):
                self.assertTrue(np.allclose(frame[1]['x'],
                                            all_data[-1][1]['x']))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Check that source modules are included in the API docs.

Typical usage is: python check_api_docs.py ../../docs/api/ ../../pyretis

"""
import pathlib
import sys


def read_api_rst(filepath):
    """Yield modules documented in the given file."""
    with filepath.open('r') as infile:
        for lines in infile:
            if lines.find('automodule') != -1:
                yield lines.strip().split()[-1]


def read_api_rst_files(dirpath):
    """Read all .rst files in the given directory."""
    rst_files = [
        i for i in dirpath.rglob('*rst')
    ]
    modules = set([])
    for filepath in rst_files:
        for mod in read_api_rst(filepath):
            modules.add(mod)
    return modules


def find_source_modules(source_dir):
    """Return source modules in the given directory."""
    source_files = set([])
    filenames = source_dir.rglob('*.py')
    for filename in filenames:
        # Figure out the file placement relative to the source
        filename = filename.relative_to(source_dir)
        # Get the stem filename
        filename_e = filename.stem
        if filename_e == '__init__':
            continue
        # Remove the .py suffix by replacing it with an empty string
        filename = filename.with_suffix("")
        mod = 'pyretis.{}'.format('.'.join(filename.parts))
        source_files.add(mod)
    return source_files


def main(docdir, source_dir):
    """Run the check."""
    docdir = pathlib.Path(docdir)
    source_dir = pathlib.Path(source_dir)
    modules = read_api_rst_files(docdir)
    source = find_source_modules(source_dir)
    missing = [i for i in source if i not in modules]
    if missing:
        print('Missing the following modules in the API documentation:')
        print()
        for i in sorted(missing):
            print(i)
        print()
        return 1
    print('No modules missing.')
    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv[1], sys.argv[2]))



#! /usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""This module will pretend to be an external program."""
import sys


if __name__ == '__main__':
    # pylint: disable=invalid-name
    print('This is a program for testing external commands.', file=sys.stdout)
    args = sys.argv
    if len(args) > 1:
        print('ERROR: Program got arguments:', file=sys.stderr, end='\n')
        argstring = ' '.join(args)
        print(argstring, file=sys.stderr, end='\n')
        sys.exit(1)
    else:
        sys.exit(0)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Define common methods and variables for the tests."""
from contextlib import contextmanager
import logging


@contextmanager
def turn_on_logging():
    """Turn on logging so that tests can detect it."""
    logging.disable(logging.NOTSET)
    try:
        yield
    finally:
        logging.disable(logging.CRITICAL)



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Methods that might be useful for testing.

This module defines methods that are useful in connection with
systems.

"""
from pyretis.core.system import System
from pyretis.core.particles import ParticlesExt


def create_system_ext(pos=None, vel=False):
    """Create an external system with given positions and velocities."""
    system = System()
    system.particles = ParticlesExt(dim=3)
    if pos is None:
        system.particles.set_pos((None, None))
    else:
        system.particles.set_pos(pos)
    system.particles.set_vel(vel)
    return system



# -*- coding: utf-8 -*-
# Copyright (c) 2026, PyRETIS Development Team.
# Distributed under the LGPLv2.1+ License. See LICENSE for more info.
"""Method to re-calculate order parameters for external trajectories.

Important methods defined here
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

recalculate_order (:py:func:`.recalculate_order`)
    Generic method for recalculating order parameters.

recalculate_from_trj (:py:func:`.recalculate_from_trj`)
    Recalculate order parameters using a GROMACS .trr file.

recalculate_from_xyz (:py:func:`.recalculate_from_xyz`)
    Recalculate order parameters using a .xyz file.

recalculate_from_frame (:py:func:`.recalculate_from_frame`)
    Recalculate order parameters using a .gro or .g96 file.

"""
import collections
import logging
import os
import numpy as np
from pyretis.core import System, ParticlesExt
from pyretis.core.box import box_matrix_to_list
from pyretis.inout import print_to_screen
from pyretis.inout.formats.gromacs import (
    read_trr_file,
    read_gromos96_file,
    read_gromacs_gro_file
)
from pyretis.inout.formats.xyz import read_xyz_file, convert_snapshot
from pyretis.inout.formats.path import PathExtFile

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.addHandler(logging.NullHandler())

__all__ = [
    'recalculate_from_trj',
    'recalculate_from_xyz',
    'recalculate_from_frame',
    'recalculate_order',
]


def recalculate_from_trj(order_parameter, trr_file, options):
    """Re-calculate order parameters from a .trr file.

    Parameters
    ----------
    order_parameter : object like :py:class:`.OrderParameter`
        The order parameter to use.
    trr_file : string
        The path to the trr file we should read.
    options: dict
        It contains:

        * `reverse`: boolean, optional
          If True, we reverse the velocities.
        * `maxidx`: integer, optional
          This is the maximum frame we will read. Can be used in case
          the .trr file contains extra frames not needed by us.
        * `minidx`: integer, optional
          This is the first frame we will read. Can be used in case we
          want to skip some frames from the .trr file.
        * `idx`: integer, optional
          This allows the selection of a single frame to recompute.

    Yields
    ------
    out : list of lists of floats
        The order parameters, calculated per frame.

    """
    system = System(box=None)  # Add dummy system.
    msg = (f'Re-calculate from {os.path.basename(trr_file)}:' +
           ' Step {}, time {}')
    minidx, maxidx = options.get('minidx'), options.get('maxidx')
    if options.get('idx', False):
        maxidx = options['idx']
        minidx = options['idx']
    for i, (header, data) in enumerate(read_trr_file(trr_file)):
        if maxidx is not None and i > maxidx:
            break
        if minidx is not None and i < minidx:
            continue
        print_to_screen(msg.format(header['step'], header['time']))
        if system.particles is None:
            system.particles = ParticlesExt(dim=data['x'].shape[1])
        system.particles.pos = data['x']
        if 'v' in data:
            if options.get('reverse', False):
                system.particles.vel = -1.0 * data['v']
            else:
                system.particles.vel = data['v']
        else:
            logger.warning('No velocities found in .trr file! Set to 0.')
            data['v'] = np.zeros_like(data['x'])
            system.particles.vel = data['v']
        length = box_matrix_to_list(data['box'])
        system.update_box(length)
        system.particles.config = (trr_file, i)
        order = order_parameter.calculate(system)
        yield order


def recalculate_from_xyz(order_parameter, traj_file, options):
    """Re-calculate order parameters from a .xyz file.

    Parameters
    ----------
    order_parameter : object like :py:class:`.OrderParameter`
        The order parameter to use.
    traj_file : string
        The path to the trajectory file we should read.
    options: dict
        It contains:

        * `reverse`: boolean, optional
          If True, we reverse the velocities.
        * `maxidx`: integer, optional
          This is the maximum frame we will read. Can be used in case
          the .trr file contains extra frames not needed by us.
        * `minidx`: integer, optional
          This is the first frame we will read. Can be used in case we
          want to skip some frames from the .trr file.
        * `box`: list of floats
          It contains the box vector lenght. It is required in the case
          that .xyz do not normally contains the simulation box dimension.

    Yields
    ------
    out : list of lists of floats
        The order parameters as a list.

    """
    system = System(box=None)
    msg = (f'Re-calculate from {os.path.basename(traj_file)}:' +
           ' Step {}')
    minidx, maxidx = options.get('minidx'), options.get('maxidx')
    reverse = options.get('reverse')

    for i, snapshot in enumerate(read_xyz_file(traj_file)):
        if maxidx is not None and i > maxidx:
            break
        if minidx is not None and i < minidx:
            continue
        print_to_screen(msg.format(i))
        box, xyz, vel, _ = convert_snapshot(snapshot)
        if box is None:
            box = options.get('box')
        if reverse:
            vel *= -1
        if system.particles is None:
            system.particles = ParticlesExt(dim=tuple(xyz.shape)[1])
        system.particles.config = (traj_file, i)
        system.particles.pos = xyz
        system.particles.vel = vel
        system.update_box(box)
        order = order_parameter.calculate(system)
        yield order


def recalculate_from_frame(order_parameter, frame_file, options):
    """Re-calculate order parameters from a .g96/.gro file.

    Here we assume that there is *ONE* frame in the ``frame_file``.

    Parameters
    ----------
    order_parameter : object like :py:class:`.OrderParameter`
        The order parameter to use.
    frame_file : string
        The path to the frame file we should read.
    options: dict
        It contains:

        * `ext`: string
          File extension for the ``frame_file``.
        * `reverse`: boolean, optional
          If True, we reverse the velocities.

    Returns
    -------
    out : list of lists of floats
        The order parameters for the current frame.

    """
    system = System(box=None)
    msg = f'Re-calculate from {os.path.basename(frame_file)}:'
    print_to_screen(msg)
    if options['ext'] == '.g96':
        _, xyz, vel, box = read_gromos96_file(frame_file)
    elif options['ext'] == '.gro':
        _, xyz, vel, box = read_gromacs_gro_file(frame_file)
    else:
        raise ValueError(f"Unknown format {options['ext']}")
    if options.get('reverse'):
        vel *= -1
    if system.particles is None:
        system.particles = ParticlesExt(dim=xyz.shape[1])
    system.particles.config = (frame_file, 0)
    system.particles.pos = xyz
    system.particles.vel = vel
    system.update_box(box)
    return [order_parameter.calculate(system)]


def recalculate_order(order_parameter, traj_file, options):
    """Re-calculate order parameters.

    Parameters
    ----------
    order_parameter : object like :py:class:`.OrderParameter`
        The order parameter to use.
    traj_file : string
        Path to the trajectory file to recalculate for.
    options: dict
        It contains:

        * `reverse`: boolean, optional
          If True, we reverse the velocities.
        * `maxidx`: integer, optional
          This is the maximum frame we will read. Can be used in case
          the .trr file contains extra frames not needed by us.
        * `minidx`: integer, optional
          This is the first frame we will read. Can be used in case we
          want to skip some frames from the .trr file.

    """
    _, ext = os.path.splitext(traj_file)

    helpers = {
        '.trr': recalculate_from_trj,
        '.xtc': recalculate_from_trj,
        '.xyz': recalculate_from_xyz,
        '.g96': recalculate_from_frame,
        '.gro': recalculate_from_frame
    }

    options['ext'] = ext
    all_order = helpers[ext](order_parameter, traj_file, options)

    if options.get('reverse'):
        return reversed(list(all_order))
    return all_order


def get_traj_files(traj_file_name):
    """Read an external file and get trajectory information.

    Parameters
    ----------
    traj_file_name : string
        The file to read trajectories from.

    Returns
    -------
    files : dict
        A dictionary with the files constituting the trajectory.

    """
    files = collections.OrderedDict()
    with PathExtFile(traj_file_name, 'r') as trajfile:
        # Just get the first trajectory:
        traj = next(trajfile.load())
        for snapshot in traj['data']:
            filename = snapshot[1]
            reverse = int(snapshot[-1]) == -1
            if filename not in files:
                files[filename] = {'minidx': None, 'maxidx': None,
                                   'reverse': reverse}
            idx = int(snapshot[2])
            minidx = files[filename]['minidx']
            if minidx is None or idx < minidx:
                files[filename]['minidx'] = idx
            maxidx = files[filename]['maxidx']
            if maxidx is None or idx > maxidx:
                files[filename]['maxidx'] = idx
    return files
